from collections import UserDict, defaultdict
from typing import TYPE_CHECKING, Any, Dict, Iterable, Optional

from .coins import COINS
from .event_delivery import EventDelivery

if TYPE_CHECKING:
    from .coin import Coin


class CustomDict:
    def __getattr__(self, name: str) -> Any:
        return self.__getitem__(name)


class ExtendedDefaultDict(defaultdict, CustomDict):
    pass


class ExtendedDict(UserDict, CustomDict):
    pass


class APIManager(EventDelivery):
    def __init__(self, wallets: Dict[str, Iterable[str]] = {}):
        super().__init__()
        self.wallets = ExtendedDefaultDict(
            lambda: ExtendedDict(),
            {currency: self.load_wallets(currency, wallets) for currency, wallets in wallets.items()},
        )
        self.event_handlers = {}

    @classmethod
    def load_wallets(cls, currency: str, wallets: Iterable[str]) -> ExtendedDict:
        return ExtendedDict({wallet: cls.load_wallet(currency, wallet) for wallet in wallets})

    @classmethod
    def load_wallet(cls, currency: str, wallet: str) -> "Coin":
        return COINS[currency](xpub=wallet)

    def add_wallet(self, currency: str, wallet: str) -> None:
        self.wallets[currency][wallet] = self.load_wallet(currency, wallet)

    def add_wallets(self, currency: str, wallets: Iterable[str]) -> None:
        self.wallets[currency].update(self.load_wallets(currency, wallets))

    def __getitem__(self, key: str) -> Any:
        return self.wallets.__getitem__(key)

    def __getattr__(self, key: str) -> Any:
        return self.__getitem__(key)

    def _merge_event_handlers(self, wallet: "Coin") -> None:
        wallet.event_handlers.update(self.event_handlers)

    def _get_event_handlers(self, wallet: "Coin") -> Iterable[str]:
        self._merge_event_handlers(wallet)
        return list(wallet.event_handlers.keys())

    async def _configure_notifications(self, autoconfigure: bool = True) -> None:
        for currency in self.wallets:
            for wallet in self.wallets[currency].values():
                try:
                    await wallet.server.subscribe(self._get_event_handlers(wallet))
                    if autoconfigure:
                        await wallet.server.configure_notifications("http://localhost:6000")
                except Exception:
                    pass

    async def process_updates(
        self, updates: Iterable[dict], currency: Optional[str] = None, wallet: Optional[str] = None
    ) -> None:
        wallet_obj = self.wallets[currency].get(wallet)
        if wallet_obj:
            self._merge_event_handlers(wallet_obj)
            await wallet_obj.process_updates(updates, pass_instance=True)



import warnings

import pytest

from bitcart import BTC
from bitcart.coin import Coin

TEST_XPUB = "zprvAWgYBBk7JR8GkUwcwyhT1qk8FQpym8cHboGyDEdMhHcL1NRe8bioZR3uo5gTSTG47iqQX6VqPL6iHHDNK55taJiV9MEscu6UiqSR1tAiSUq"
REGTEST_XPUB = (
    "vprv9DMUxX4ShgxMMQduKMSnoNsX4jZjdqmRRapGRRFbok1XXrjkvFyAnvSVPbs4t8ZDA73fTT9DLzdCyHBh39AZHG8nsP1gEj11EwSdYP8zhKF"
)
LIGHTNING_UNSUPPORTED_XPUB = (
    "xprv9s21ZrQH143K3tZPHG8CbfZ7uUY5stdHmaEXeSqawGrZuAoBdHPgKHjdkfmHSdxDJSbo29JiU1PcWhzEsgFryqMHQfr2T5TWBPK8EqFjscZ"
)


@pytest.fixture
async def xpub():
    return TEST_XPUB


@pytest.fixture
async def coin():
    return Coin()


@pytest.fixture
async def btc_nowallet():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        return BTC()


@pytest.fixture
async def btc_wallet():
    return BTC(xpub=TEST_XPUB)


@pytest.fixture
async def lightning_unsupported_wallet():
    return BTC(xpub=LIGHTNING_UNSUPPORTED_XPUB)


@pytest.fixture
async def regtest_wallet():
    return BTC(xpub=REGTEST_XPUB)


@pytest.fixture
async def regtest_node_id():
    return await BTC(xpub=REGTEST_XPUB, rpc_url="http://localhost:5110").node_id


@pytest.fixture
async def btc():
    with warnings.catch_warnings():  # to ignore no xpub passed warning
        warnings.simplefilter("ignore")
        btc_obj = BTC()
    return btc_obj



import multiprocessing
import time

import pytest
from aiohttp import ClientSession

test_queue = multiprocessing.Queue()


@pytest.yield_fixture
def setup_webhook(btc_wallet):
    btc_wallet.add_event_handler("new_transaction", new_tx_handler)
    process = multiprocessing.Process(target=btc_wallet.start_webhook)  # wallet required
    process.start()
    time.sleep(2)
    yield
    process.terminate()
    process.join()


def new_tx_handler(event, tx):
    test_queue.put(True)


@pytest.mark.asyncio
async def test_event_delivery(setup_webhook):
    async with ClientSession() as session:
        async with session.post("http://localhost:6000") as resp:  # no json passed, silently ignoring
            assert await resp.json() == {}
        assert test_queue.qsize() == 0
        async with session.post(
            "http://localhost:6000", json={"updates": [{"event": "new_transaction", "tx": "test"}]}
        ) as resp:
            assert await resp.json() == {}
        assert test_queue.qsize() == 1
        assert test_queue.get() is True



import multiprocessing
import time

import pytest
from aiohttp import ClientSession

from bitcart import BTC, GZRO, LTC
from bitcart.manager import APIManager

pytestmark = pytest.mark.asyncio

test_queue = multiprocessing.Queue()


def new_tx_handler(instance, event, tx):
    if isinstance(instance, BTC):
        test_queue.put(True)


@pytest.fixture
async def manager(xpub):
    return APIManager({"BTC": [xpub, xpub]})


@pytest.fixture
async def webhook_manager(xpub):
    return APIManager({"BTC": [xpub, "test"]})


@pytest.yield_fixture
def setup_webhook(webhook_manager):
    webhook_manager.add_event_handler("new_transaction", new_tx_handler)
    process = multiprocessing.Process(target=webhook_manager.start_webhook)
    process.start()
    time.sleep(2)
    yield
    process.terminate()
    process.join()


async def test_manager_storage(manager, xpub):
    assert manager.wallets == {"BTC": {xpub: BTC(xpub=xpub)}}
    assert manager.wallets.BTC == manager.wallets["BTC"] == manager.BTC == manager["BTC"]


async def test_manager_classmethods(xpub):
    assert APIManager.load_wallet("BTC", xpub) == BTC(xpub=xpub)
    assert APIManager.load_wallets("BTC", [xpub, xpub]) == {xpub: BTC(xpub=xpub)}


async def test_manager_add_wallets(manager, xpub):
    assert manager.add_wallet("LTC", xpub) is None
    assert manager.wallets == {"BTC": {xpub: BTC(xpub=xpub)}, "LTC": {xpub: LTC(xpub=xpub)}}
    assert manager.add_wallets("GZRO", [xpub, xpub]) is None
    assert manager.wallets == {"BTC": {xpub: BTC(xpub=xpub)}, "LTC": {xpub: LTC(xpub=xpub)}, "GZRO": {xpub: GZRO(xpub=xpub)}}


async def test_manager_start_webhook(setup_webhook, xpub):
    async with ClientSession() as session:
        async with session.post(
            "http://localhost:6000", json={"wallet": xpub, "updates": [{"event": "new_transaction", "tx": "test"}]}
        ) as resp:
            assert await resp.json() == {}
        assert test_queue.qsize() == 1
        assert test_queue.get() is True



import asyncio
import logging
from json import JSONDecodeError
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, Optional, Union

from aiohttp import web

if TYPE_CHECKING:
    from providers.jsonrpcrequests import RPCProxy


class EventDelivery:
    server: "RPCProxy"
    event_handlers: Dict[str, Callable]

    async def handle_webhook(self, request: "web.Request") -> "web.Response":
        try:
            json = await request.json()
        except JSONDecodeError:
            return web.json_response({})
        await self.process_updates(json.get("updates", []), json.get("currency", "BTC"), json.get("wallet"))
        return web.json_response({})

    async def process_updates(
        self, updates: Iterable[dict], currency: Optional[str] = None, wallet: Optional[str] = None
    ) -> None:
        raise NotImplementedError()

    def _configure_webhook(self) -> None:
        self.webhook_app = web.Application()
        self.webhook_app.router.add_post("/", self.handle_webhook)

    async def _configure_notifications(self, autoconfigure: bool = True) -> None:
        await self.server.subscribe(list(self.event_handlers.keys()))
        if autoconfigure:
            await self.server.configure_notifications("http://localhost:6000")

    async def configure_webhook(self, autoconfigure: bool = True) -> None:
        self._configure_webhook()
        await self._configure_notifications(autoconfigure=autoconfigure)

    def _start_webhook(self, port: int = 6000, **kwargs: Any) -> None:
        web.run_app(self.webhook_app, port=port, **kwargs)

    def start_webhook(self, port: int = 6000, **kwargs: Any) -> None:
        self.configure_webhook()
        self._start_webhook(port=port, **kwargs)

    async def poll_updates(self, timeout: Union[int, float] = 1) -> None:  # pragma: no cover
        """Poll updates

        Poll daemon for new transactions in wallet,
        this will block forever in while True loop checking for new transactions

        Example can be found on main page of docs

        Args:
            self (BTC): self
            timeout (Union[int, float], optional): seconds to wait before requesting transactions again. Defaults to 1.

        Raises:
            InvalidEventError: If server sent invalid event name not matching ALLOWED_EVENTS

        Returns:
            None: This function runs forever
        """
        await self.server.subscribe(list(self.event_handlers.keys()))
        while True:
            try:
                data = await self.server.get_updates()
            except Exception as err:
                logging.error(err)
                await asyncio.sleep(timeout)
                continue
            await self.process_updates(data)
            await asyncio.sleep(timeout)

    def add_event_handler(self, events: Union[Iterable[str], str], func: Callable) -> None:
        """Add event handler to handle event(s) provided

        Args:
            self (BTC): self
            events (Union[Iterable[str], str]): event or events
            func (Callable): function to handle those

        Returns:
            None: None
        """
        if isinstance(events, str):
            events = [events]
        for event in events:
            self.event_handlers[event] = func

    def on(self, events: Union[Iterable[str], str]) -> Callable:
        """Register on event

        Register callback function to be run when event is emmited

        All available events are accessable as:

        >>> btc.ALLOWED_EVENTS
        ['new_block', 'new_transaction']

        Function signature must be

        .. code-block:: python

            def handler(event, **kwargs):

        kwargs sent differ from event to event, as for now
        new_block event sends height kwarg as new block height
        new_transaction event sends tx kwarg as tx_hash of new transaction

        Args:
            self (BTC): self
            events (Union[Iterable[str], str]): event name or list of events for function to be run on

        Returns:
            Callable: It is a decorator
        """

        def wrapper(f: Callable) -> Callable:
            self.add_event_handler(events, f)
            return f

        return wrapper
